﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Eliza
{
    /// <summary>
    /// A rule that associates decomposition of an input string with a collection of
    /// reassembled responses.
    /// </summary>
    class DecompReassemblyRule
    {
        #region Constructor

        /// <summary>
        /// Initializes a new instance of the <see cref="DecompReassemblyRule"/> class.
        /// </summary>
        /// <param name="decomp">How the input string should be decomposed.</param>
        /// <param name="reassembledResponses">The reassembled responses.</param>
        public DecompReassemblyRule
            (string decomp,
             List<string> reassembledResponses)
        {
            this._decomp = decomp;
            this._reassembledResponses = new List<string>();
            foreach(string reassembledResponse in reassembledResponses) {
                this._reassembledResponses.Add(reassembledResponse);
            }
            this._nextResponseIndex = 0;
        }

        #endregion

        #region Public methods

        /// <summary>
        /// Determines whether this instance can decompose the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns>
        /// <see langword="true"/> if this instance can decompose the specified input; otherwise,
        /// <see langword="false"/>.
        /// </returns>
        public bool CanDecompose
            (string input)
        {
            // Pad the input with spaces
            input = string.Format(" {0} ", input.Trim());

            // Check if the input contains the phrase being searched for - if not, return false
            string phraseSearchedFor = this._decomp.Replace("*", string.Empty);
            phraseSearchedFor = phraseSearchedFor.Trim();
            phraseSearchedFor = string.Format(" {0} ", phraseSearchedFor);
            int phraseIndex = input.IndexOf(phraseSearchedFor);
            if (phraseIndex == -1) {
                return false;
            }

            // Otherwise, extract the text element that occurs before the phrase
            string elementBeforePhrase = input.Substring(0, phraseIndex);
            elementBeforePhrase = elementBeforePhrase.Trim();

            // And the text element that occurs after the phrase
            int indexOfEndOfPhrase = phraseIndex + phraseSearchedFor.Length;
            string elementAfterPhrase = input.Substring(indexOfEndOfPhrase, input.Length - indexOfEndOfPhrase);
            elementAfterPhrase = elementAfterPhrase.Trim();

            // Store the elements for later use by GenerateResponse()
            this._textElements = new List<string>();
            this._textElements.Add(elementBeforePhrase);
            this._textElements.Add(elementAfterPhrase);

            // We found the phrase, so return true
            return true;
        }

        /// <summary>
        /// Generates a response to the user's input.
        /// </summary>
        /// <param name="input">The user's input.</param>
        /// <returns>
        /// The response.
        /// </returns>
        public string GenerateResponse
            (string input)
        {
            // Select the next reponse
            string response = this._reassembledResponses[this._nextResponseIndex];
            this._nextResponseIndex++;
            if (this._nextResponseIndex == this._reassembledResponses.Count) {
                this._nextResponseIndex = 0;
            }

            // Replace any numbered placeholders with the corresponding text elements, after
            // transforming them
            for (int i=0; (i < this._textElements.Count); i++) {
                string placeHolder = string.Format("({0})", i + 1);
                if (response.Contains(placeHolder)) {
                    string transformedInputElement = StringTransformer.PostTransform(this._textElements[i]);
                    response = response.Replace(placeHolder, transformedInputElement);
                }
            }

            // Reduce double spaces with a single space and remove spaces before punctuation
            while(response.Contains("  ")) {
                response = response.Replace("  ", " ");
            }
            response = response.Replace(" ?", "?");
            response = response.Replace(" !", "!");
            response = response.Replace(" .", ".");
            response = response.Replace(" ,", ",");
            response = response.Trim();

            // Return the response
            return response;
        }

        #endregion

        #region Fields

        /// <summary>
        /// How the input string should be decomposed.
        /// </summary>
        private string _decomp;

        /// <summary>
        /// The text elements in the user's input.
        /// </summary>
        private List<string> _textElements;

        /// <summary>
        /// The collection of reassembled responses.
        /// </summary>
        private List<string> _reassembledResponses;

        /// <summary>
        /// Index of the next reassembled response to return.
        /// </summary>
        private int _nextResponseIndex;

        #endregion
    }
}
